/*****************************************************************************
 * pce                                                                       *
 *****************************************************************************/

/*****************************************************************************
 * File name:     contrib/mkimg/mkdosemu.c                                   *
 * Created:       2003-04-19 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-01-03 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 1999-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id$ */


/*****************************************************************************
DOSEMU image file format:

offset size description
0      7    magic ("DOSEMU\0")
7      4    number of heads
11     4    sectors per track
15     4    number of tracks
19     4    file pointer to image start

- numbers are little endian
- the image seems to start at 128 for all existing files
******************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef HAVE_FTRUNCATE
#include <unistd.h>
#include <sys/types.h>
#endif


#include "mbr.h"


static
void buf_set_uint32 (void *buf, unsigned long i, unsigned long val)
{
  unsigned char *tmp;

  tmp = buf;
  tmp += i;

  tmp[0] = val & 0xff;
  tmp[1] = (val >> 8) & 0xff;
  tmp[2] = (val >> 16) & 0xff;
  tmp[3] = (val >> 24) & 0xff;
}

int main (int argc, char **argv)
{
  unsigned long c, h, s, n;
  unsigned char buf[4096];

  if (argc != 4) {
    fprintf (stderr, "usage: %s cylinders heads sectors\n", argv[0]);
    return (1);
  }

  c = strtoul (argv[1], NULL, 0);
  h = strtoul (argv[2], NULL, 0);
  s = strtoul (argv[3], NULL, 0);

  if ((c == 0) || (h == 0) || (s == 0)) {
    fprintf (stderr, "%s: bad c/h/s\n", argv[0]);
    return (1);
  }

  memset (buf, 0, 128);
  memcpy (buf, "DOSEMU\x00", 7);
  buf_set_uint32 (buf, 7, h);
  buf_set_uint32 (buf, 11, s);
  buf_set_uint32 (buf, 15, c);
  buf_set_uint32 (buf, 19, 128);
  fwrite (buf, 1, 128, stdout);

  mbr[510] = 0x55;
  mbr[511] = 0xaa;
  fwrite (mbr, 1, 512, stdout);

  n = c * h * s - 1;

#ifdef HAVE_FTRUNCATE
  if (ftruncate (fileno (stdout), 128 + 512 + 512 * (off_t) n) == 0) {
    return (0);
  }
#endif

  memset (buf, 0, 4096);

  while (n >= 8) {
    fwrite (buf, 8, 512, stdout);
    n -= 8;
  }

  while (n > 0) {
    fwrite (buf, 1, 512, stdout);
    n -= 1;
  }

  return (0);
}
