/*****************************************************************************
 * pce                                                                       *
 *****************************************************************************/

/*****************************************************************************
 * File name:     contrib/mkimg/mkpce.c                                      *
 * Created:       2005-01-01 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-11-29 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005 Hampa Hug <hampa@hampa.ch>                        *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id$ */


/*
PCE image file format

0    4    Magic (PIMG)
4    4    version (0)
8    4    image offset
12   4    block count
16   4    cylinders
20   4    heads
24   4    sectors
28   4    block size
*/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>

#ifdef HAVE_FTRUNCATE
#include <unistd.h>
#include <sys/types.h>
#endif


#define DSK_PCE_MAGIC 0x50494d47UL


#include "mbr.h"


static
void set_uint32_be (void *buf, unsigned i, uint32_t val)
{
  unsigned char *tmp = (unsigned char *) buf + i;

  tmp[0] = (val >> 24) & 0xff;
  tmp[1] = (val >> 16) & 0xff;
  tmp[2] = (val >> 8) & 0xff;
  tmp[3] = val & 0xff;
}

int main (int argc, char **argv)
{
  uint32_t c, h, s, n;
  unsigned char buf[4096];

  if (argc != 4) {
    fprintf (stderr, "usage: %s cylinders heads sectors\n", argv[0]);
    return (1);
  }

  c = strtoul (argv[1], NULL, 0);
  h = strtoul (argv[2], NULL, 0);
  s = strtoul (argv[3], NULL, 0);
  n = c * h * s;

  if (n == 0) {
    fprintf (stderr, "%s: bad c/h/s\n", argv[0]);
    return (1);
  }

  memset (buf, 0, 128);
  set_uint32_be (buf, 0, DSK_PCE_MAGIC);
  set_uint32_be (buf, 4, 0);
  set_uint32_be (buf, 8, 128);
  set_uint32_be (buf, 12, n);
  set_uint32_be (buf, 16, c);
  set_uint32_be (buf, 20, h);
  set_uint32_be (buf, 24, s);
  set_uint32_be (buf, 28, 512);
  fwrite (buf, 1, 128, stdout);

  mbr[510] = 0x55;
  mbr[511] = 0xaa;
  fwrite (mbr, 512, 1, stdout);

  if (n <= 1) {
    return (0);
  }

#ifdef HAVE_FTRUNCATE
  if (ftruncate (fileno (stdout), 128 + 512 * (off_t) n) == 0) {
    return (0);
  }
#endif

  memset (buf, 0, 4096);

  if (fseeko (stdout, 128 + 512 * (off_t) (n - 1), SEEK_SET) == 0) {
    if (fwrite (buf, 512, 1, stdout) != 1) {
      return (1);
    }

    return (0);
  }

  while (n >= 8) {
    if (fwrite (buf, 512, 8, stdout) != 8) {
      return (1);
    }

    n -= 8;
  }

  while (n > 0) {
    if (fwrite (buf, 512, 1, stdout) != 1) {
      return (1);
    }

    n -= 1;
  }

  return (0);
}
